import asyncio
from contextlib import asynccontextmanager
from dataclasses import dataclass
from decimal import Decimal
from typing import AsyncIterator, NamedTuple, Optional, Sequence, Union

import aiohttp
import aiohttp.web
import pytest
from yarl import URL


@dataclass
class ApiAddress:
    host: str
    port: int


@dataclass(frozen=True)
class AdminChargeRequest:
    idempotency_key: Optional[str]
    cluster_name: str
    username: str
    amount: Decimal


@dataclass(frozen=True)
class AdminDebtRequest:
    idempotency_key: Optional[str]
    cluster_name: str
    username: str
    amount: Decimal


class AdminServer(NamedTuple):
    address: ApiAddress
    app: aiohttp.web.Application

    @property
    def url(self) -> URL:
        return URL(f"http://{self.address.host}:{self.address.port}/api/v1/")

    @property
    def requests(self) -> Sequence[Union[AdminChargeRequest, AdminDebtRequest]]:
        return [request for request in self.app["requests"]]


@pytest.fixture
async def mock_admin_server(
    loop: asyncio.AbstractEventLoop,
) -> AsyncIterator[AdminServer]:
    async def _handle_quota_patch(request: aiohttp.web.Request) -> aiohttp.web.Response:
        cluster_name = request.match_info["cname"]
        username = request.match_info["uname"]
        payload = await request.json()
        amount = Decimal(payload["additional_quota"]["credits"])
        idempotency_key = request.query.get("idempotency_key")
        app["requests"].append(
            AdminChargeRequest(
                idempotency_key=idempotency_key,
                cluster_name=cluster_name,
                username=username,
                amount=amount,
            )
        )
        return aiohttp.web.Response()

    async def _handle_add_debt(request: aiohttp.web.Request) -> aiohttp.web.Response:
        cluster_name = request.match_info["cname"]
        payload = await request.json()
        username = payload["user_name"]
        amount = Decimal(payload["credits"])
        idempotency_key = request.query.get("idempotency_key")
        app["requests"].append(
            AdminDebtRequest(
                idempotency_key=idempotency_key,
                cluster_name=cluster_name,
                username=username,
                amount=amount,
            )
        )
        return aiohttp.web.Response()

    def _create_app() -> aiohttp.web.Application:
        app = aiohttp.web.Application()
        app["requests"] = []
        app.router.add_routes(
            (
                aiohttp.web.patch(
                    "/api/v1/clusters/{cname}/users/{uname}/quota", _handle_quota_patch
                ),
                aiohttp.web.post("/api/v1/clusters/{cname}/debts", _handle_add_debt),
            )
        )
        return app

    app = _create_app()
    runner = ApiRunner(app, port=8085)
    api_address = await runner.run()
    yield AdminServer(address=api_address, app=app)
    await runner.close()


@pytest.fixture
def admin_url(
    mock_admin_server: AdminServer,
) -> URL:
    return mock_admin_server.url


@asynccontextmanager
async def create_local_app_server(
    app: aiohttp.web.Application, port: int = 8080
) -> AsyncIterator[ApiAddress]:
    runner = aiohttp.web.AppRunner(app)
    try:
        await runner.setup()
        api_address = ApiAddress("0.0.0.0", port)
        site = aiohttp.web.TCPSite(runner, api_address.host, api_address.port)
        await site.start()
        yield api_address
    finally:
        await runner.shutdown()
        await runner.cleanup()


class ApiRunner:
    def __init__(self, app: aiohttp.web.Application, port: int) -> None:
        self._app = app
        self._port = port

        self._api_address_future: asyncio.Future[ApiAddress] = asyncio.Future()
        self._cleanup_future: asyncio.Future[None] = asyncio.Future()
        self._task: Optional[asyncio.Task[None]] = None

    async def _run(self) -> None:
        async with create_local_app_server(self._app, port=self._port) as api_address:
            self._api_address_future.set_result(api_address)
            await self._cleanup_future

    async def run(self) -> ApiAddress:
        loop = asyncio.get_event_loop()
        self._task = loop.create_task(self._run())
        return await self._api_address_future

    async def close(self) -> None:
        if self._task:
            task = self._task
            self._task = None
            self._cleanup_future.set_result(None)
            await task

    @property
    def closed(self) -> bool:
        return not bool(self._task)



from setuptools import find_packages, setup


setup(
    name="neuro-admin-client",
    python_requires=">=3.8",
    url="https://github.com/neuro-inc/neuro-admin-client",
    packages=find_packages(),
    setup_requires=["setuptools_scm"],
    install_requires=[
        "aiohttp>=3.7",
    ],
    use_scm_version=True,
    include_package_data=True,
)



from decimal import Decimal

from neuro_admin_client import AdminClient

from .conftest import AdminChargeRequest, AdminDebtRequest, AdminServer


class TestAdminClient:
    async def test_charge_user_credits_null(self) -> None:
        cluster_name = "test-cluster"
        username = "username"
        amount = Decimal("20.11")
        key = "key"
        async with AdminClient(base_url=None) as client:
            await client.change_user_credits(cluster_name, username, amount, key)

        assert client._client is None

    async def test_add_debt_null(self) -> None:
        cluster_name = "test-cluster"
        username = "username"
        amount = Decimal("20.11")
        key = "key"
        async with AdminClient(base_url=None) as client:
            await client.add_debt(cluster_name, username, amount, key)

        assert client._client is None

    async def test_charge_user_credits(self, mock_admin_server: AdminServer) -> None:
        cluster_name = "test-cluster"
        username = "username"
        amount = Decimal("20.11")
        key = "key"
        async with AdminClient(base_url=mock_admin_server.url) as client:
            await client.change_user_credits(cluster_name, username, amount, key)

        assert len(mock_admin_server.requests) == 1
        assert mock_admin_server.requests[0] == AdminChargeRequest(
            key,
            cluster_name,
            username,
            amount,
        )

    async def test_add_debt(self, mock_admin_server: AdminServer) -> None:
        cluster_name = "test-cluster"
        username = "username"
        amount = Decimal("20.11")
        key = "key"
        async with AdminClient(base_url=mock_admin_server.url) as client:
            await client.add_debt(cluster_name, username, amount, key)

        assert len(mock_admin_server.requests) == 1
        assert mock_admin_server.requests[0] == AdminDebtRequest(
            key,
            cluster_name,
            username,
            amount,
        )



from contextlib import asynccontextmanager
from decimal import Decimal
from typing import Any, AsyncIterator, Optional, Sequence

import aiohttp
from multidict import CIMultiDict
from yarl import URL


class AdminClient:
    def __init__(
        self,
        *,
        base_url: Optional[URL],
        service_token: Optional[str] = None,
        conn_timeout_s: int = 300,
        read_timeout_s: int = 100,
        conn_pool_size: int = 100,
        trace_configs: Sequence[aiohttp.TraceConfig] = (),
    ):
        self._base_url = base_url
        self._service_token = service_token
        self._conn_timeout_s = conn_timeout_s
        self._read_timeout_s = read_timeout_s
        self._conn_pool_size = conn_pool_size
        self._trace_configs = trace_configs
        self._client: Optional[aiohttp.ClientSession] = None

    async def __aenter__(self) -> "AdminClient":
        self._init()
        return self

    async def __aexit__(self, *args: Any) -> None:
        await self.aclose()

    async def aclose(self) -> None:
        if not self._client:
            return
        await self._client.close()
        del self._client

    def _init(self) -> None:
        if self._client:
            return
        if self._base_url is None:
            return
        connector = aiohttp.TCPConnector(limit=self._conn_pool_size)
        timeout = aiohttp.ClientTimeout(
            connect=self._conn_timeout_s, total=self._read_timeout_s
        )
        self._client = aiohttp.ClientSession(
            headers=self._generate_headers(self._service_token),
            connector=connector,
            timeout=timeout,
            trace_configs=list(self._trace_configs),
        )

    def _generate_headers(self, token: Optional[str] = None) -> "CIMultiDict[str]":
        headers: "CIMultiDict[str]" = CIMultiDict()
        if token:
            headers["Authorization"] = f"Bearer {token}"
        return headers

    @asynccontextmanager
    async def _request(
        self, method: str, path: str, **kwargs: Any
    ) -> AsyncIterator[aiohttp.ClientResponse]:
        assert self._client
        assert self._base_url is not None
        url = self._base_url / path
        async with self._client.request(method, url, **kwargs) as response:
            response.raise_for_status()
            yield response

    async def change_user_credits(
        self,
        cluster_name: str,
        username: str,
        credits_delta: Decimal,
        idempotency_key: str,
    ) -> None:
        if self._base_url is None:
            return
        payload = {"additional_quota": {"credits": str(credits_delta)}}
        async with self._request(
            "PATCH",
            f"clusters/{cluster_name}/users/{username}/quota",
            json=payload,
            params={"idempotency_key": idempotency_key},
        ) as response:
            response.raise_for_status()

    async def add_debt(
        self,
        cluster_name: str,
        username: str,
        credits: Decimal,
        idempotency_key: str,
    ) -> None:
        if self._base_url is None:
            return
        payload = {"user_name": username, "credits": str(credits)}
        async with self._request(
            "POST",
            f"clusters/{cluster_name}/debts",
            json=payload,
            params={"idempotency_key": idempotency_key},
        ) as response:
            response.raise_for_status()
